% OptimalMBCApprox
% Apply optimal MBC approximation to CART3 data.

% Load Data
load cart3_reg3_10states_14mps

% Indices for states, inputs, and outputs:
%   q = [Tower 1st F/A; Rotor Pos; Blade1 1st Flap; Blade2; Blade3]
%   x = [q; qdot]
%   u = [Blade1 Pitch; Blade2; Blade3; Gen Torque];
%   y = [Rotor Speed; Blade1 Root Moment; Blade2; Blade3];
% This ordering corresponds to the paper.
% Note that the data has the disturbances before the control inputs.
ndof = 5;
nx = 2*ndof;
nu = 4;
idxu = [3 4 5 2];
ny = 4;
idxy = 1:4;

% Grab state matrices for PLTV system
A = AMat;
B = BMat(:,idxu,:);
C = CMat(idxy,:,:);
D = DMat(idxy,idxu,:);

% MBC tranxformation matrices
psi = 0:2:358;
Np = length(psi);

M = zeros(3,3,Np);
Minv = M;

Mu = zeros(nu,nu,Np);
Muinv = Mu;

My = zeros(ny,ny,Np);
Myinv = My;

Mx = zeros(nx,nx,Np);
Mxinv = Mx;
Mxdot = zeros(nx,nx,Np);
for i=1:Np
    % Rotor Angle
    p = psi(i)*pi/180;
    
    % Single MBC matrix
    Mi = [1 sin(p) cos(p); ...
        1 sin(p+2*pi/3) cos(p+2*pi/3); ...
        1 sin(p+4*pi/3) cos(p+4*pi/3)];
    
    % Flip columns to correspond to MBC3 code
    Mi = Mi(:,[1 3 2]);
    
    % Store M and its inverse
    M(:,:,i) = Mi;
    Minv(:,:,i) = inv( Mi );

    % Derivative of Mi
    Midot = Omega(i)*[0 cos(p) -sin(p); ...
        0 cos(p+2*pi/3) -sin(p+2*pi/3); ...
        0 cos(p+4*pi/3) -sin(p+4*pi/3)];
    Midot = Midot(:,[1 3 2]);
    
    % Second Derivative of Mi (Assuming Omegadd = 0)
    Midd = Omega(i)^2*[0 -sin(p) -cos(p); ...
        0 -sin(p+2*pi/3) -cos(p+2*pi/3); ...
        0 -sin(p+4*pi/3) -cos(p+4*pi/3)];
    Midd = Midd(:,[1 3 2]);
    
    % I/O Tranxformationx
    Mu(:,:,i) = blkdiag(Mi,1);
    Muinv(:,:,i) = inv( Mu(:,:,i) );
    My(:,:,i) = blkdiag(1,Mi);
    Myinv(:,:,i) = inv( My(:,:,i) );
    
    % State Tranxformation
    Mq = blkdiag(eye(2),Mi);
    Mqdot = blkdiag(zeros(2),Midot);
    Mqdd = blkdiag(zeros(2),Midd);
    Mx(:,:,i) = [ Mq zeros(ndof); Mqdot Mq];
    Mxinv(:,:,i) = inv(Mx(:,:,i));
    Mxdot(:,:,i) = [ Mqdot zeros(ndof); Mqdd Mqdot];    
end

% Method 1: Average
A1 = mean(A,3);
B1 = mean(B,3);
C1 = mean(C,3);
D1 = mean(D,3);
%D1 - AvgDMat(idxy,idxu)

% Method 2: MBC + Average
A2tmp = zeros(nx,nx,Np);
B2tmp = zeros(nx,nu,Np);
C2tmp = zeros(ny,nx,Np);
D2tmp = zeros(ny,nu,Np);
for i=1:Np
    A2tmp(:,:,i) = Mxinv(:,:,i)*( A(:,:,i)*Mx(:,:,i) - Mxdot(:,:,i) );
    B2tmp(:,:,i) = Mxinv(:,:,i)*B(:,:,i)*Mu(:,:,i);
    C2tmp(:,:,i) = Myinv(:,:,i)*C(:,:,i)*Mx(:,:,i);    
    D2tmp(:,:,i) = Myinv(:,:,i)*D(:,:,i)*Mu(:,:,i);
end
A2 = mean(A2tmp,3);
B2 = mean(B2tmp,3);
C2 = mean(C2tmp,3);
D2 = mean(D2tmp,3);
%D2 - MBC_AvgD(idxy,idxu)

% Method 3: Convex Opt + MBC
bb = []; AA = [];
for i=1:Np
    tmp = A(:,:,i)-Mxdot(:,:,i)*Mxinv(:,:,i);
    bb = [bb; tmp(:)];    
    tmp = kron(Mxinv(:,:,i)',Mx(:,:,i));
    AA = [AA; tmp];
end
A3 = AA\bb;
A3 = reshape(A3,[nx nx]);


bb = []; AA = [];
for i=1:Np
    tmp = B(:,:,i);
    bb = [bb; tmp(:)];    
    tmp = kron(Muinv(:,:,i)',Mx(:,:,i));
    AA = [AA; tmp];
end
B3 = AA\bb;
B3 = reshape(B3,[nx nu]);

bb = []; AA = [];
for i=1:Np
    tmp = C(:,:,i);
    bb = [bb; tmp(:)];    
    tmp = kron(Mxinv(:,:,i)',My(:,:,i));
    AA = [AA; tmp];
end
C3 = AA\bb;
C3 = reshape(C3,[ny nx]);

bb = []; AA = [];
for i=1:Np
    tmp = D(:,:,i);
    bb = [bb; tmp(:)];    
    tmp = kron(Muinv(:,:,i)',My(:,:,i));
    AA = [AA; tmp];
end
D3 = AA\bb;
D3 = reshape(D3,[ny nu]);

% Comparison of approximation:
% Map each LTI approximation back to an PLTV system via the MBC
% and compare with the original system.
D1M = zeros(ny,nu,Np);
D2M = zeros(ny,nu,Np);
D3M = zeros(ny,nu,Np);
for i=1:Np
   D1M(:,:,i) = My(:,:,i)*D1*Muinv(:,:,i);
   D2M(:,:,i) = My(:,:,i)*D2*Muinv(:,:,i);
   D3M(:,:,i) = My(:,:,i)*D3*Muinv(:,:,i);    
end

E1 = D-D1M;
E2 = D-D2M;
E3 = D-D3M;

% I think the two entries below should be equal
[norm(E3(:)), norm(AA*D3(:)-bb)]



